#! /usr/bin/perl
#-*- tab-width: 4; -*-
# ex:ts=4
#
# portlint - lint for port directory
#
# this version implemented by:
#   Michael Haro <mharo@FreeBSD.org>
# versions thru 2.x implemented by:
#	Jun-ichiro itojun Hagino <itojun@itojun.org>
#	Yoshishige Arai <ryo2@on.rim.or.jp>
#
# Copyright(c) 2000,2001 by Michael Haro <mharo@FreeBSD.org>.
# Copyright(c) 1997 by Jun-ichiro Hagino <itojun@itojun.org>.
# All rights reserved.
# Freely redistributable.  Absolutely no warranty.
#
# $FreeBSD: ports/devel/portlint/src/portlint.pl,v 1.15 2000/03/20 18:03:01 mharo Exp $
# $Id: portlint.pl,v 1.57 2001/08/18 07:27:45 mharo Exp $
#

# My current idea is to have a bunch of really small
# functions that each do one (1) check and have main
# call all these functions on whatever files they operate on

#
# creating a new portlint check
# -----------------------------
# - create a seperate function to do the check in
# - the function should call &verbose to report no errors
#
# the reason we use seperate functions for everything is to make it 
# easy to disable indivisual checks with a command line option
# (there is a reason we are using Getopt::Long)
#
# creating a new set of functions (seperate file)
# -------------------------------
# return an array of the following hash structure (one element per function)
# {
#   function => \&function # pointer to function
#   name => "Name",        # English name for check
#   option => "option"     # Commmand line options which enable/disable check",
#   enabled => 1,    # 1 = enabled, 0 = disabled
#   file => "Makefile",    # file to check
# }

use strict;
use vars qw/%makevar %mastersites %opts $nfatal $nwarn $ninfo $prog/;
use Getopt::Long;
use LWP;

#
# Constants
#
sub enabled		{ 1 }
sub disabled	{ 0 }
sub default		{ undef }


#
# Exported Global Variables
#

# my name
($prog) = ($0 =~ /([^\/]+)$/);

# Command line options hash
%opts = ();
%makevar = ();

# number of warnings and fatal errors
$ninfo = 0;
$nwarn = 0;
$nfatal = 0;

#
# Private Global Variables (this file only)
#

# version variables
my $major = 3;
my $minor = 0;
my $patchlevel = 0;
my $version = "$major.$minor.$patchlevel";

my $patchcount = 0;
my %checkedurls;

my @func_reg = ();

######################################## LOAD SCRIPTS

my $plprefix = "/home/mharo/work/src/portlint/share/portlint/";

# checks that apply to every section
push(@func_reg, @{require "$plprefix/checkall.pl"});

# Makefile checks
push(@func_reg, @{require "$plprefix/makefile.pl"});
push(@func_reg, @{require "$plprefix/makefilesections.pl"});
require "$plprefix/makefile2.pl";
# portlint 2's checkmakefile (unconverted code)
require "$plprefix/old_checkmakefile.pl";

# pkg-plist checks
push(@func_reg, @{require "$plprefix/checkplist.pl"});

# pkg-descr and pkg-comment checks
push(@func_reg, @{require "$plprefix/checkpkg.pl"});

# other checks
push(@func_reg, @{require "$plprefix/checkother.pl"});

######################################## START LINTING

&main;

errx(1, "got to end of main, but didn't exit");

######################################## OTHER STUFF

# put main into a subroutine to take care of random global vars
sub main {
	my @varnames = ();
	my @sectvars = ();

	Getopt::Long::Configure("no_auto_abbrev", "no_getopt_compat",
							"bundling", "ignore_case");
	my $result = GetOptions(\%opts,
							"b",
							"checkall|a",
							"contblank|B=i",
							"committer|c",
							"patch",
							"patchfile=s",
							"display-makevars!",
							"display-options!",
							"help|h|H",
							"newport|N",
							"make_vars|M",
							"tabs|t",
							"warnings-fatal",
							"version|V",
							"verbose|v",
							);

	&usage if $result eq '' || $opts{help};
	&version if ($opts{version});

# load portlintrc hash and merge it with the %opts hash
	if (-r "$ENV{HOME}/.portlintrc") {
		my %rchash = %{require "$ENV{HOME}/.portlintrc"};
		foreach my $key (%rchash) {
			next if ($rchash{$key} == undef); #ignore "default" case
			if ($rchash{$key} != &enabled && $rchash{$key} != &disabled) {
				print STDERR "\$HOME/.portlintrc: $key: unknown value, ignoring it.\n";
				next;
			}
			$opts{$key} = $rchash{$key};
		}
	}
	
	my $portdir = $ARGV[0] ? $ARGV[0] : ".";
	&verbose("linting directory $portdir");
	chdir($portdir) || &err(1, "chdir $portdir");

# display-options if verbose and nodisplay-options is not defined
# or if display-options is defined
	if ((defined($opts{'display-options'}) && $opts{'display-options'} != 0 && $opts{'verbose'}) || ($opts{'verbose'} && !defined($opts{'display-options'})) || $opts{'display-options'} ) {
		foreach my $var (sort keys %opts) {
			print "option: $var = $opts{$var}\n";
		}
	}

# init stuff
	my $contblank = $opts{contblank} ? $opts{contblank} : 1;

my @varlist = qw(
	PORTNAME PORTVERSION PORTREVISION PORTEPOCH PKGNAME PKGNAMEPREFIX
	PKGNAMESUFFIX DISTNAME DISTFILES CATEGORIES MASTERDIR MAINTAINER
	WRKDIR WRKSRC NO_WRKSUBDIR PATCHDIR SCRIPTDIR FILESDIR PKGDIR
	COMMENT DESCR PLIST MD5_FILE .CURDIR VALID_CATEGORIES NO_CHECKSUM
	MAN1 MAN2 MAN3 MAN4 MAN5 MAN6 MAN7 MAN8 MAN9 MANL MANN
	MASTER_SITES PREFIX PKGINSTALL PKGDEINSTALL PKGREQ PKGMESSAGE
	INSTALLS_SHLIB PORTSDIR
);

	%makevar = &getmakevars(@varlist);

# Read bsd.sites.mk
	my $sites_mk = "$makevar{PORTSDIR}/Mk/bsd.sites.mk";
	open(MK, $sites_mk) || die "$sites_mk: $!";
	my @site_groups = grep($_ = /^(MASTER_SITE_\w+)/ && $1, <MK>);
	close(MK);
	%mastersites = &getmakevars(@site_groups);

	my $pwd = $makevar{".CURDIR"};

# remove these in the future
	my $file;
	my ($whole, $linecnt, $linelen);

# Makefile checks
	$file = "Makefile";
	&verbose("checking $file");
	($whole, $linecnt, $linelen) = &loadfile($file);

	&functionloop($file, $whole);

	&blanklines($file, $whole, $contblank);
	&nouservar($file, "NO_CHECKSUM", $makevar{NO_CHECKSUM});
	# any other vars that users define?

	# direct use of command names check
	# full paths check
	
	# check for duplicate var defs
	# check that MANX= foo.Y  has X == Y
	# check for duplicate manpage listings in MANX
	&propermansect(%makevar);
	# check for MASTER_SITES which are defined in bsd.port.mk

	# old portlint2 functions
	print "Old checkmakefile begin\n";
	&checkmakefile($file, \%makevar);
	print "Old checkmakefile end\n";

#MICHAEL:  add something to deal with master/slave/independent ports


# pkg-descr checks
	$file = "pkg-descr";
	$file = $makevar{DESCR} if $makevar{DESCR} ne "";
	$file =~ s,^$pwd/,,;
	&verbose("checking $file");
	($whole, $linecnt, $linelen) = &loadfile($file);

	&functionloop($file, $whole);

	&maxlines($file, $linecnt, 24);
	&maxchars($file, $linelen, 80);
	&blanklines($file, $whole, $contblank);

# pkg-comment checks
	$file = "pkg-comment";
	$file = $makevar{COMMENT} if $makevar{COMMENT} ne "";
	$file =~ s,^$pwd/,,;
	&verbose("checking $file");
	($whole, $linecnt, $linelen) = &loadfile($file);

	&functionloop($file, $whole);

	&maxlines($file, $linecnt, 1);
	&maxchars($file, $linelen, 70);

# pkg-plist checks
	$file = "pkg-plist";
	$file = $makevar{PLIST} if $makevar{PLIST} ne "";
	$file =~ s,^$pwd/,,;
	&verbose("checking $file");
	($whole, $linecnt, $linelen) = &loadfile($file);

	&functionloop($file, $whole);

	&blanklines($file, $whole, 0);
	# plist @exec/@unexec install-info ordering check
	# deal with multiple PLIST files

# distinfo checks
	$file = "distinfo";
	$file = $makevar{MD5_FILE} if $makevar{MD5_FILE} ne "";
	$file =~ s,^$pwd/,,;
	&verbose("checking $file");
	($whole, $linecnt, $linelen) = &loadfile($file);

	&functionloop($file, $whole);

# scripts/ related checks
	# &lastline foreach file in scripts/

# other checks
	&workdircheck;
	&warncrud(".");

# We are done.  Display number of warnings and errors
	if ($nfatal || $nwarn) {
		print "$nfatal fatal errors and $nwarn warnings found.\n";
	} else {
		print "looks fine.\n";
	}
	
	exit $nfatal;
}

# eventually usage should be copied into the portlint manpage
# so for now this might be a little too verbose
sub usage {
	print STDERR <<EOF;
usage: $prog [options] [port_directory]
    -B#, --contblank#  allow # contiguous blank lines (default: 1 line)
    -b  warn \$(VARIABLE)
    -t, --tabs         nit pick about use of spaces.
    -v, --verbose      verbose mode
    --warnings-fatal   treat warnings as fatals.
    --patch            create a portlint generated patchfile with suggested
                       changes which can be applied to the port.
    --patchfile=name   change the name of the portlint generated patch file.

put something here about how the following options can be disabled with 'no'
    --display-makevars display variables taken from make.
                       defaults to on when in verbose mode.  May
                       be disabled by --nodisplay-options
    --display-options  show which command line options were seen.
                       defaults to on when in verbose mode.  May
                       be disabled by --nodisplay-options

unimplemented portlint 2.2.2 flags:
    -a  additional check for scripts/* and pkg/*
    -c  committer mode
    -M  set make variables (ex. PORTSDIR=/usr/ports.work)
    -N  writing a new port
EOF
		exit 0;
}

sub version {
	print "$prog version $version\n";
	exit 0;
}

sub functionloop {
	my ($file, $whole) = @_;

	foreach my $item (@func_reg) {
		if ($item->{file} eq "$file" || $item->{file} eq undef) {
			if ($item->{enabled}) {  ## add command line options stuff here
				#print "running check $item->{name} on file $file\n";
				&{$item->{function}}($file, $whole);
			}
		}
	}
}

#
# user notification functions
#
sub perror {
	my ($msg) = @_;

	warnx("perror called w/o WARN or FATAL or INFO") if ($msg !~ /^(WARN|FATAL|INFO):/);

	$msg =~ s/^WARN/FATAL/ if ($opts{'warnings-fatal'});
	$nfatal++ if ($msg =~ /^FATAL/);
	$nwarn++ if ($msg =~ /^WARN/);
	$ninfo++ if ($msg =~ /^INFO/);
	$msg =~ s/^WARN:/WARN: /;
	$msg =~ s/^INFO:/INFO: /;

	my @text = split(//, $msg);
	print "len: $#text\n";

# line wrap the portlint error message
#MICHAEL: I think this code has a bug in it somewhere but it is too late at
#         night for me to run it thru my brain
	my @arr = split(//, $msg);
	my $i = 0;
	my $start = 0;
	while ($#arr >= $i) {
		for ($i = 75 + $start; $i > $start && ($arr[$i] ne " " && $i < $#arr); $i--) {;}
		print " " x 7 if ($start != 0);
		print join(//, @arr[$start..$i]) . "\n";
		if ($start == 0) {
			$start = $i + 1;
		} else {
			$start = $i + 1 + 7;
		}
	}

}

sub verbose {
	my ($msg) = @_;

	print "GOOD: $msg\n" if ($opts{verbose});
}

#
# functions to deal with internal errors
#
sub warn {
	my ($msg) = @_;
		print STDERR $prog . ": " . $msg . "$!\n";
}

sub warnx {
		my ($msg) = @_;
		print STDERR $prog . ": " . $msg . "\n";
}

sub err {
		my ($ex, $msg) = @_;
		warnx("WARNING: err called incorrectly") if (($ex !~ m/^\d+/) || ($msg eq ""));
		print STDERR $prog . ": " . $msg . ": $!\n";
		exit $ex;
}

sub errx {
		my ($ex,$msg) = @_;
		warnx("WARNING: errx called incorrectly") if (($ex !~ m/^\d+/) || ($msg eq ""));
		print STDERR $prog . ": " . $msg . "\n";
		exit $ex;
}

#
# portlint general functions
#
sub addlist {
	my ($varnames, $varlist) = @_;

	foreach my $item (@$varlist) {
		push(@$varnames, $item);
	}
}

sub getmakevars {
	my (@varlist) = @_;
	my %hash;
	my $cmd = join(' -V ', "make -DPORTLINTING $opts{makeenv}".
		" MASTER_SITE_BACKUP=''", @varlist);

	my $i = 0;
	for (split(/\n/, `$cmd`)) {
		$hash{$varlist[$i]} = $_;
		if ((defined($opts{'display-makevars'}) && $opts{'display-makevars'} != 0 && $opts{'verbose'}) || ($opts{'verbose'} && !defined($opts{'display-makevars'})) || $opts{'display-makevars'} ) {
			print "makevar: $varlist[$i] = $_\n";
		}
		$i++;
	}

	return %hash;
}

sub loadfile {
	my ($file) = @_;
	my $linecnt = 0;
	my $maxlinelen = 0;
	my $len;
	my $whole = "";
	
	if (!open(F, "< $file")) {
		&perror("FATAL: unable to open $file");
		return ("", 0, 0);
	}
	while(<F>) {
		$whole .= $_;
		$linecnt++;
		$len = length(chomp($_));
		$maxlinelen = $len if ($len > $maxlinelen);
	}
	close(F);

	if ($whole =~ /[\033\200-\377]/) {
		&perror("WARN: $file: including iso-8859-1, or ".
			"other local characters.  Should be a plain ascii file");
	}

	return ($whole, $linecnt, $maxlinelen);
}

sub contains {
	my ($search, @list) = @_;

	for (@list) {
		return 1 if m/^\Q$search\E$/;
	}
	return 0;
}

sub checkearlier {
	my ($file, $whole, @varnames) = @_;

	my $i;
	foreach my $i (@varnames) {
		if ($whole =~ /\n$i[?+]?=/) {
			&perror("WARN: \"$i\" has to appear earlier in $file.");
		}
	}
}

# this function was just a test and should NOT be used.
# it probably is totally broken too
sub createpatch {
	my ($file, $pattern, $old, $new) = @_;

	return if ($opts{'patch'} ne 1);
	my $patchfile = "portlint.patch";
	$patchfile = $opts{'patchfile'} if $opts{'patchfile'} ne "";
	$patchfile = "$patchfile-$patchcount";

#MICHAEL: not sure if patch creation is correct
	my ($oldline, $oldlen, $newline, $newlen);
	my $oldout;
	my $newout;

	# stupid hack that I don't believe is the right way to do this
	my $beginning = 0;
	$pattern =~ s/^(\^)//;
	$beginning = 1 if $1 ne "";

	open(F, "$file") || die "$!";
	while(<F>) {
		if ($beginning) {
			last if (m/(^$pattern)/);
		} else {
			last if (m/($pattern)/);
		}
	}
	$oldline = $.;
	$newline = $.;
	if ($1 eq "") {
		warnx("unable to match pattern \'$pattern\' in file $file");
		return;
	}
	close(F);
	foreach my $line (split(/\n/, $old)) {
		$oldout .= "- $line\n";
		$oldlen++;
	}
	foreach my $line (split(/\n/, $new)) {
		$newout .= "+ $line\n";
		$newlen++;
	}

	errx(1, "\`$patchfile\' file already exists.  I can't handle this yet") if ( -e "$patchfile");
	open(DIFF, ">>$patchfile");
	print DIFF <<EOF;
--- $file
+++ $file
@@ -$oldline +$newline,$newlen @@
EOF
	print DIFF "$oldout$newout";
}

sub geturlstatus {
	my ($url) = @_;

	return if $checkedurls{"$url"} == 1;

	my $ua = new LWP::UserAgent;
	my $request = new HTTP::Request('GET', $url);
	my $response = $ua->request($request);

	my ($statuscode, $statusmessage) = split(/ /, $response->status_line, 2);

	if ($statuscode != 200) {
		&perror("WARN: URL \'$url\' returned error $statuscode, $statusmessage");
	}
	$checkedurls{"$url"} = 1;
}

sub maxlines {
	my ($file, $count, $max) = @_;

	if ($count > $max) {
		&perror("WARN: $file exceeds $max lines, make it shorter if possible. (currently $count lines)");
	} else {
		&verbose("$file has $count line(s).");
	}
}

sub maxchars {
	my ($file, $count, $max) = @_;

	if ($count > $max) {
		&perror("WARN: $file: includes lines that exceed $max characters.");
	} else {
		&verbose("$file has at most $max characters per line.");
	}
}

#MICHAEL: maybe include the line number this is found on
sub blanklines {
	my ($file, $whole, $contblank) = @_;
	
	my $blanks = "\n" x ($contblank + 2);

	if ($whole =~ /$blanks/) {
		&perror("FATAL: $file: contiguous blank lines (> $contblank lines) found");
	} else {
		&verbose("$file has no contiguous blank lines");
	}
}
