#
# pkg-plist related checks
# $Id: checkplist.pl,v 1.4 2001/04/30 20:42:07 mharo Exp $
#

use strict;
use vars qw/%makevar %mastersites %opts $nfatal $nwarn $ninfo $prog/;
use Data::Dumper;

########################################  FUNCTIONS BELOW THIS POINT

sub plistfullpath {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^\//) {
			&perror("FATAL: $file $i: use of full pathname is not allowed.");
			$problem = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file doesn't use full paths");
	}
}

sub plistlibtool {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /\.la$/) {
			&perror("WARN: $file $i: installing libtool archives, ".
				"please use USE_LIBTOOL in Makefile if possible");
			$problem = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file doesn't include .la files");
	}
}

sub plistinfodir {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^info\/dir$/) {
			&perror("FATAL: $file $i: use install-info to add/remove an ".
				"entry from \"info/dir\", don't install info/dir");
			$problem = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file no info/dir file found");
	}
}

sub plistcommands {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^\@/) {
			if ($line =~ /^\@(cwd|cd)[ \t]+(\S+)/) {
				&perror("WARN: $file $i: use of \@$1 is discouraged");
				$problem=1;
			} elsif ($line =~ /^\@unexec[ \t]+rmdir/ && $line !~ /true$/) {
				&perror("WARN: $file $i: use \"\@dirrm\" instead of \"\@unexec rmdir\".");
				$problem=1;
			} elsif ($line =~ /^\@(owner|group)\s/) {
				&perror("WARN: $file $i: \@$1 should not be needed in PLIST");
				$problem=1;
			} elsif ($line =~ /^\@(dirrm|option|exec|unexec|comment)/) {
				; # no check made
			} else {
				&perror("WARN: $file $i: ".
					"unknown PLIST directive \"$line\"");
			}
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file has no package-list @ command problems");
	}
}

sub plistinfofile {
	my ($file, $whole) = @_;

	my @exec_info = ();
	my @unexec_info = ();
	my @infofiles = ();

	my $problem = 0;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^\@exec[ \t]+install-info\s+(\%D\/)?(.+)\s+.+$/) {
			push(@exec_info, $2);
		} elsif ($line =~ /^\@unexec[ \t]+install-info[ \t]+--delete\s+(\%D\/)?(.+)\s+(.+)$/) {
			push(@unexec_info, $2);
		} elsif ($line =~ /^info\/.*info$/) {
			push(@infofiles, $line);
		}
	}

# check that every infofile has an exec install-info and unexec install-info
	foreach my $if (@infofiles) {
		if (! &contains($if, @exec_info)) {
			&perror("FATAL: $file: you need an '\@exec install-info \%D/$if \%D/info/dir' line");
			$problem = 1;
		}
		if (! &contains($if, @unexec_info)) {
			&perror("FATAL: $file: you need an '\@unexec install-info --delete \%D/$if \%D/info/dir' line");
			$problem = 1;
		}
	}
	if (! $problem) {
		&verbose("$file has no problems with info files");
	}
}

sub plistman {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ m#man/([^/]+/)?man(.)/([^\.]+\.(.))(\.gz)?$#) {
			my $mansect = $2;
			my $manpage = $3;
			my $manpagesect = $4;
			my $mancompress = $5;
			&perror("WARN: $file $i: $manpage should be listed in Makefile, not PLIST");
			if ($mansect != $manpagesect) {
				&perror("FATAL: $file $i: $manpage is being installed into section $mansect but lists itself as section $manpagesect.");
			}
		}
	}
	if (! $problem) {
		&verbose("$file has no manpages listed in the PLIST");
	}
}


########################################  FUNCTION REGISTRATION

my @func_reg = (
{
  function => \&plistfullpath,
  name => "plistfullpath",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
{
  function => \&plistlibtool,
  name => "plistlibtool",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
{
  function => \&plistinfodir,
  name => "plistinfodir",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
{
  function => \&plistcommands,
  name => "plistcommands",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
{
  function => \&plistinfofile,
  name => "plistinfofile",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
{
  function => \&plistman,
  name => "plistman",
  option => undef,
  enabled => 1,
  file => "pkg-plist"
},
);

#print Data::Dumper->Dump([\@func_reg],['@func_reg']);
\@func_reg;
