#
# makefile related checks
# $Id: makefile.pl,v 1.6 2001/04/30 20:42:07 mharo Exp $
#

use strict;
use vars qw/%makevar %mastersites %opts $nfatal $nwarn $ninfo $prog/;

use LWP;
use Data::Dumper;


########################################  FUNCTIONS BELOW THIS POINT

sub checkspaceindent {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^        /) {  # 8 spaces
			&perror("WARN: $file $i: use a tab (not a space) to indent");
			$problem = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file has no space indentation problems");
	}
}

sub tabaftervariable {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /(\w+.?)= /) {
#MICHAEL: is 11 the right number?
			if (length($1) < 11) {
				&perror("WARN: $file $i: use a tab (not a space) after a variable name");
				$problem = 1;
			}
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file has no spaces after variable declarations");
	}
}

sub checkparens {
	my ($file, $whole) = @_;

	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /\$\(([\w\d]+)\)/) {
			&perror("WARN: $file $i: use \${$1} instead of \$($1)");
			$problem = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file uses '{' and '}' around all variable names");
	}
}

sub dupvarcheck {
	my ($file, $whole) = @_;

	my %seenvar = ();
	my $problem = 0;
	my $i = 1;
	foreach my $line (split(/\n/, $whole)) {
		if ($line =~ /^([\w\d]+)(.)?=/) {
			if ($seenvar{$1} != 1 && defined($2)) {
				&perror("INFO: $file $i: variable $1 initially defined ".
					"by '$2='");
				$problem = 1;
			}
			if ($seenvar{$1} == 1 && !defined($2)) {
				&perror("FATAL: $file $i: variable $1 is redefined.");
				$problem = 1;
			}
			$seenvar{$1} = 1;
		}
		$i++;
	}
	if (! $problem) {
		&verbose("$file has no duplicate variable definations");
	}
}

sub pkgname {
	my ($file, $whole) = @_;
	
	if ($whole =~ m/\nPKGNAME\s*=/) {
		&perror("FATAL: $file: PKGNAME is obsoleted by PORTNAME, ".
			"PORTVERSION, PKGNAMEPREFIX and PKGNAMESUFFIX.");
	} else {
		&verbose("PKGNAME not found in $file");
	}
}

sub gzipflags {
	my ($file, $whole) = @_;

	if ($whole =~ m/\${GZIP_CMD}\s+-(\w+(\s+-)?)*(\d)/) {
		&perror("WARN: possible use of \'\${GZIP_CMD} -$3\' ".
			"found. \${GZIP_CMD} includes \'-\${GZIP}\' which ".
			"sets the compression level.");
	} else {
		&verbose("no compression level found after \${GZIP_CMD}");
	}
}

sub mkdirflags {
	my ($file, $whole) = @_;

	if ($whole =~ m/\${MKDIR}\s+-p/) {
		&perror("WARN: possible use of \'\${MKDIR} -p\' ".
			"found. \${MKDIR} includes \'-p\' by default.");
	} else {
		&verbose("\${MKDIR} -p not found");
	}
}

#MICHAEL: should probably use make vars and not parse Makefile for this
#MICHAEL: not tested
sub is_interactive {
	my ($file, $whole) = @_;

	if ($whole =~ m/^IS_INTERACTIVE.?\s*=/m) {
		if ($whole !~ m/^(BATCH|FOR_CDROM).?\s*=/m) {
			&perror("WARN: use of IS_INTERACTIVE discouraged. provide batch mode by using BATCH and/or FOR_CDROM.");
		}
	} else {
		&verbose("$file: IS_INTERACTIVE not found");
	}
}

#MICHAEL: not tested, just "ported" from portlint2 so not even checked
#         for validity
sub abspathname {
	my($file, $whole) = @_;
	my($s, $i, %cmdnames);
	my($pre);

	# ignore parameter string to echo command
	$whole =~ s/[ \t][\@-]?(echo|\$[\{\(]ECHO[\}\)]|\$[\{\(]ECHO_MSG[\}\)])[ \t]+("(\\'|\\"|[^"])*"|'(\\'|\\"|[^"])*')[ \t]*[;\n]//;

	&verbose("OK: checking direct use of full pathnames in $file.");
	foreach my $s (split(/\n+/, $whole)) {
		$i = '';
		if ($s =~ /(^|[ \t\@'"-])(\/[\w\d])/) {
			# suspected pathnames are recorded.
			$i = $2 . $';
			$pre = $` . $1;

			if ($pre =~ /MASTER_SITE_SUBDIR/) {
				# MASTER_SITE_SUBDIR lines are ok.
				$i = '';
			}
		}
		if ($i ne '') {
			$i =~ s/\s.*$//;
			$i =~ s/['"].*$//;
			$i = substr($i, 0, 20) . '...' if (20 < length($i));
			&perror("WARN: possible use of absolute pathname ".
				"\"$i\", in $file.") unless ($i =~ m,^/compat/,);
		}
	}

	&verbose("OK: checking direct use of pathnames, phase 1.");
#MICHAEL: shouldn't hard code these paths
%cmdnames = split(/\n|\t+/, <<EOF);
/usr/ports	\${PORTSDIR} instead
/usr/local	\${PREFIX} or \${LOCALBASE}, as appropriate
/usr/X11R6	\${PREFIX} or \${X11BASE}, as appropriate
EOF
	foreach my $i (keys %cmdnames) {
		if ($whole =~ /$i/) {
			&perror("WARN: possible direct use of \"$&\" ".
				"found in $file. if so, use $cmdnames{$i}.");
		}
	}

	&verbose("OK: checking direct use of pathnames, phase 2.");
%cmdnames = split(/\n|\t+/, <<EOF);
distfiles	\${DISTDIR} instead
pkg			\${PKGDIR} instead
files		\${FILESDIR} instead
scripts		\${SCRIPTDIR} instead
patches		\${PATCHDIR} instead
work		\${WRKDIR} instead
EOF
	foreach my $i (keys %cmdnames) {
		if ($whole =~ /(\.\/|\$[\{\(]\.CURDIR[\}\)]\/|[ \t])(\b$i)\//) {
			&perror("WARN: possible direct use of \"$i\" ".
				"found in $file. if so, use $cmdnames{$i}.");
		}
	}
}

sub noportdocs {
	my ($file, $whole) = @_;

	if ($whole =~ m/\${PREFIX}\/share\// && $whole !~ m/^\.if \(?\!defined\(NOPORTDOCS\)/m) {
		&perror("WARN: use \".if !defined(NOPORTDOCS)\" to wrap installation of files into \${PREFIX}/share/doc.");
	} else {
		&verbose("any files being installed into \${PREFIX} are fine");
	}
}

########################################  FUNCTION REGISTRATION

my @func_reg = (
{
  function => \&checkspaceindent,
  name => "checkspaceindent",
  option => "checkspaceindent",
  enabled => 1,
  file => "Makefile"
},
{
  function => \&tabaftervariable,
  name => "tabaftervariable",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&checkparens,
  name => "checkparens",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&dupvarcheck,
  name => "dupvarcheck",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&pkgname,
  name => "pkgname",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&gzipflags,
  name => "gzipflags",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&mkdirflags,
  name => "mkdirflags",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&is_interactive,
  name => "is_interactive",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&abspathname,
  name => "abspathname",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
{
  function => \&noportdocs,
  name => "noportdocs",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
);

#print Data::Dumper->Dump([\@func_reg],['@func_reg']);
\@func_reg;
