#
# makefile sections (gets around problem with current function registration code
# $Id: makefilesections.pl,v 1.4 2001/04/30 20:42:07 mharo Exp $
#

use strict;
use vars qw/%makevar %mastersites %opts $nfatal $nwarn $ninfo $prog/;

sub sectioncheck {
	my ($file, $whole) = @_;

	my @varnames = ();

	# deal with Makefile sections
	my @sections = &sectionsplit($whole);
	my $idx = 0;

	# section 0 - comment lines
	&commentsonly($file, $sections[$idx]);
	&sect0lines($file, $sections[$idx]);
	$idx++;

	# section 1 - PORTNAME/PORTVERSION
	&commentsstrip($sections[$idx]);
	my @sectvars = split(/\s+/, <<EOF);
PORTNAME PORTVERSION PORTREVISION PORTEPOCH CATEGORIES MASTER_SITES
MASTER_SITE_SUBDIR PKGNAMEPREFIX PKGNAMESUFFIX DISTNAME EXTRACT_SUFX
DISTFILES DIST_SUBDIR EXTRACT_ONLY
EOF
	&checkorder($file, 'PORTNAME', $sections[$idx], @sectvars);
	# verify PORTNAME and PORTVERSION are set and valid
	# sanity check PORTNAME/PORTVERSION
	# check CATEGORIES is set
	# committer check - modules alias name warning
	&validcategories($makevar{VALID_CATEGORIES}, $makevar{CATEGORIES});
	# various CATEGORIES warnings / checks
	# check mastersites URLs
	# get makefile definations for variables and compare with make -V results
	# check for bogus EXTRACT_SUFX
	# DISTFILES check
	
	&addlist(\@varnames, \@sectvars);
	$idx++;

	# section 2 - PATCH_SITES/PATCHFILES (optional)
	@sectvars = split(/\s+/, <<EOF);
PATCH_SITES PATCH_SITE_SUBDIR PATCHFILES PATCH_DIST_STRIP
EOF
	&commentsstrip($sections[$idx]);
	# check if we have a section 2
	my $have_section2 = 0;
	foreach my $item (@sectvars) {
		if ($sections[$idx] =~ /$item/) {
			$have_section2 = 1;
		}
	}
	if ($have_section2) {
		&checkearlier($file, $sections[$idx], @varnames);
		&checkorder($file, 'PATCH_SITES', $sections[$idx], @sectvars);
		#remove checked lines from $section[$idx] before calling checkextra
		# thats how the current implementation (from portlint2 does it)
		#&checkextra($file, $sections[$idx], 'PATCH_SITES');
		&addlist(\@varnames, \@sectvars);
		$idx++;
	}

	# section 3 - MAINTAINER
	@sectvars = split(/\s+/, <<EOF);
MAINTAINER
EOF
	&commentsstrip($sections[$idx]);
	&checkearlier($file, $sections[$idx], @varnames);
	&maintainerexists($file, $sections[$idx], $makevar{MAINTAINER});
	&maintainervalidemail($file, $makevar{MAINTAINER});
	&maintaineremailappend($file, $sections[$idx]);
	&addlist(\@varnames, \@sectvars);
	$idx++;

	# section 4 - *_DEPENDS (optional)
	&commentsstrip($sections[$idx]);
	$idx++;

	# section 5 - rest (optional)
	@sectvars = split(/\s+/, <<EOF);
RESTRICTED NO_CDROM NO_PACKAGE NO_CONFIGURE NO_PATCH
EOF
	my $rest = join("\n\n", @sections[$idx .. $#sections]);
	&commentsstrip($rest);
	&checkearlier($file, $rest, @varnames);
	&restrictedcheck($file, $rest);
	&nocdromcheck($file, $rest);
	&nopackagecheck($file, $rest);
	&noconfigurecheck($file, $rest);
	&nopatchcheck($file, $rest);
}

######################################## Makefile Generic

sub sectionsplit {
	my ($whole) = @_;

	# keep comment, blank line, comment in the same section
	$whole =~ s/(#.*\n)\n+(#.*)/$1$2/g;
	my @sections = split(/\n\n+/, $whole);
	for (my $i = 0; $i <= $#sections; $i++) {
		if ($sections[$i] !~ /\n$/) {
			$sections[$i] .= "\n";
		}
	}

	return @sections;
}

# Warning: this function modifies the argument passed
sub commentsstrip {
	$_[0] =~ s/^#.*$//mg;
	$_[0] =~ s/\n+/\n/g;
}

#MICHAEL: not tested
sub checkorder {
	my ($file, $section, $whole, @order) = @_;

	my @items = ();
	foreach my $i (split(/\n/, $whole)) {
		$i =~ s/[+?]?=.*$//m;
		push(@items, $i) if ($i =~ m/^\w/);
	}
	
	my $j = -1;
	my $invalidorder = 0;
	while ($#items >= 0) {
		my $i = shift(@items);
		my $k = 0;
		while ($k <= $#order && $order[$k] ne $i) {
			$k++;
		}
		if ($order[$k] eq $i) {
			if ($k < $j) {
				&perror("FATAL: $i appears out-of-order.");
				$invalidorder++;
			} else {
				&verbose("OK: seen $i, in order.");
			}
			$j = $k;
		} else {
			&perror("FATAL: $file: extra item \"$i\" placed in the section $section.");
		}
	}
	if ($invalidorder) {
		&perror("FATAL: order must be " . join('/', @order) . '.');
	} else {
		&verbose("OK: $section section is ordered properly.");
	}
}

sub checkearlier {
	my ($file, $whole, @varnames) = @_;

	my $i;
	foreach my $i (@varnames) {
		if ($whole =~ /\n$i[?+]?=/) {
			&perror("WARN: \"$i\" has to appear earlier in $file.");
		}
	}
}

#MICHAEL: not tested, just "ported" from portlint2 so not even checked
#         for validity
sub checkextra {
	my ($file, $whole, $section) = @_;

	$whole = "\n" . $whole if ($whole !~ /^\n/);
	$whole =~ s/\n#[^\n]*/\n/g;
	$whole =~ s/\n\n+/\n/g;
	$whole =~ s/^\s+//;
	$whole =~ s/\s+$//;
	return if ($whole eq '');

	if ($whole =~ /^([\w\d]+)/) {
		&perror("WARN: $file: extra item placed in the $section section, ".
			"for example, \"$1\".");
	} else {
		&perror("WARN: $file: extra item placed in the $section section.");
	}
}


######################################## Makefile Section 0

sub commentsonly {
	my ($file, $whole) = @_;
	
	if ($whole =~ m/^[^#]/m) {
		&perror("FATAL: non-comment line in comment section of $file.");
	} else {
		&verbose("only comments found in comments section.");
	}
}

#MICHAEL: should probably check the order of the lines
sub sect0lines {
	my ($file, $whole) = @_;

	my @linestocheck = split(/\n/, <<EOF);
New Ports Collection Makefile For
Whom
Date Created
EOF

	foreach my $item (@linestocheck) {
		if ($whole !~ m/^#\s*$item:\s+\S/mi) {
			&perror("FATAL: no '$item' line in comment section of $file.");
		} else {
			&verbose("'$item' line found");
		}
	}

	if ($whole =~ m/^Version Required/im) {
		&perror("WARN: $file: Version required is no longer needed in the comment section.");
	}

	if ($whole !~ m/^#(\s+)\$FreeBSD([^\$]*)\$$/m) {
		&perror("FATAL: $file: no \$FreeBSD\$ line in comment section.");
	} else {
		if ($1 ne " ") {
			 &perror("WARN: $file: please use single whitespace right before ".
				"\$FreeBSD\$ tag.");
		}
		if ($2 ne "" && $opts{newport}) {
			&perror("WARN: $file: for new port, make \$FreeBSD\$ tag in ".
				"comment section empty, to make CVS happy.");
		} elsif ($2 eq "" && $opts{newport} != 1) {
			&perror("WARN: $file: \$FreeBSD\$ tag in comment section should ".
				"not be empty");
		}
	}
}

######################################## Makefile Section 1

sub validcategories {
	my ($validcats, $catlist) = @_;
	foreach my $cat (split(/\s+/, $catlist)) {
		if ($validcats !~ m/$cat/) {
			&perror("FATAL: Makefile: $cat is not a valid category");
		} else {
			&verbose("$cat is a valid category");
		}
	}
}

######################################## Makefile Section 2
######################################## Makefile Section 3

sub maintainerexists {
	my ($file, $whole, $makevar) = @_;
	
	if ($whole !~ m/^MAINTAINER[?]?=/m && ($makevar eq "" || $makevar eq "ports\@FreeBSD.org")) {
		&perror("FATAL: no MAINTAINER listed in $file.");
	} else {
		&verbose("maintainer found.");
	}
}

sub maintainervalidemail {
	my ($file, $makevar) = @_;
	
	if ($makevar !~ m/[A-z0-9]+\@[A-z0-9]+\.[A-z0-9]+/) {
		&perror("WARN: maintainer email address, '$makevar', does not appear to be valid.");
	} else {
		&verbose("maintainer appears to have a valid email address");
	}
}

sub maintaineremailappend {
	my ($file, $whole) = @_;
	if ($whole =~ m/^MAINTAINER\+=/m) {
		&perror("FATAL: MAINTAINER can not be assigned using +=.");
	} else {
		&verbose("maintainer email not appended.");
	}
}

######################################## Makefile Section 4
######################################## Makefile Section 5

sub restrictedcheck {
	my ($file, $whole) = @_;
	if ($whole =~ m/^RESTRICTED[+?]?=/m) {
		&perror("WARN: \"RESTRICTED\" found. do not forget to update ports/LEGAL");
	} else {
		&verbose("RESTRICTED not found in $file.");
	}
}

sub nocdromcheck {
	my ($file, $whole) = @_;
	if ($whole =~ m/^NO_CDROM[+?]?=/m) {
		&perror("WARN: \"NO_CDROM\" found. do not forget to update ports/LEGAL");
	} else {
		&verbose("NO_CDROM not found in $file.");
	}
}

sub nopackagecheck {
	my ($file, $whole) = @_;
	if ($whole =~ m/^NO_PACKAGE[+?]?=/m) {
		&perror("WARN: \"NO_PACKAGE\" found. do not forget to update ports/LEGAL");
	} else {
		&verbose("NO_PACKAGE not found in $file.");
	}
}

sub noconfigurecheck {
	my ($file, $whole) = @_;
	if ($whole =~ m/^NO_CONFIGURE[+?]?=/m) {
		&perror("FATAL: \"NO_CONFIGURE\" is obsolete. remove this.");
	} else {
		&verbose("NO_CONFIGURE not found in $file.");
	}
}

sub nopatchcheck {
	my ($file, $whole) = @_;
	if ($whole =~ m/^NO_PATCH[+?]?=/m) {
		&perror("FATAL: \"NO_PATCH\" is obsolete. remove this.");
	} else {
		&verbose("NO_PATCH not found in $file.");
	}
}

########################################  FUNCTION REGISTRATION

my @func_reg = (
{
  function => \&sectioncheck,
  name => "sectioncheck",
  option => undef,
  enabled => 1,
  file => "Makefile"
},
);

#print Data::Dumper->Dump([\@func_reg],['@func_reg']);
\@func_reg;
